@@ -3,6 +3,35 @@
  
 #include "fitsio.h"
 
+/* 
+    Threading support using POSIX threads programming interface
+    (supplied by Bruce O'Neel) 
+
+    All threaded programs MUST have the 
+
+    -D_REENTRANT
+
+    on the compile line and must link with -lpthread.  This means that
+    when one builds cfitsio for threads you must have -D_REENTRANT on the
+    gcc or cc command line.
+*/
+
+#ifdef _REENTRANT
+#include <pthread.h>
+#include <assert.h>
+extern pthread_mutex_t Fitsio_Lock;
+extern int Fitsio_Pthread_Status;
+
+#define FFLOCK1(lockname)   (assert(!(Fitsio_Pthread_Status = pthread_mutex_lock(&lockname))))
+#define FFUNLOCK1(lockname) (assert(!(Fitsio_Pthread_Status = pthread_mutex_unlock(&lockname))))
+#define FFLOCK   FFLOCK1(Fitsio_Lock)
+#define FFUNLOCK FFUNLOCK1(Fitsio_Lock)
+
+#else
+#define FFLOCK
+#define FFUNLOCK
+#endif
+
 /*
   If REPLACE_LINKS is defined, then whenever CFITSIO fails to open
   a file with write access because it is a soft link to a file that
@@ -19,13 +48,9 @@
 
 #define DBUFFSIZE 28800 /* size of data buffer in bytes */
 
-#define NIOBUF  40  /* number of IO buffers to create (default = 40) */
-         /* !! Significantly increasing NIOBUF may degrade performance !! */
 #define NMAXFILES  300   /* maximum number of FITS files that can be opened */
         /* CFITSIO will allocate (NMAXFILES * 80) bytes of memory */
 
-#define IOBUFLEN 2880    /* size in bytes of each IO buffer (DONT CHANGE!) */
-
 #define MINDIRECT 8640   /* minimum size for direct reads and writes */
                          /* MINDIRECT must have a value >= 8640 */
 
@@ -54,7 +79,7 @@
 #define BYTESWAPPED TRUE
 #define LONGSIZE 64
 
-#elif defined(__sparcv9)
+#elif defined(__sparcv9) || (defined(__sparc__) && defined(__arch64__))
                                /*  SUN Solaris7 in 64-bit mode */
 #define BYTESWAPPED FALSE
 #define MACHINE NATIVE
@@ -145,6 +170,12 @@
 
 #define BYTESWAPPED TRUE
  
+#elif defined(__tile__)
+
+/*  64-core 8x8-architecture Tile64 platform */
+
+#define BYTESWAPPED TRUE
+ 
 #else
 
 /*  assume all other machine uses the same IEEE formats as used in FITS files */
@@ -261,8 +292,9 @@
 
 
 #define COMPRESS_NULL_VALUE -2147483647
+#define N_RANDOM 10000  /* DO NOT CHANGE THIS;  used when quantizing real numbers */
 
-int ffmkky(char *keyname, char *keyval, char *comm, char *card, int *status);
+int ffmkky(const char *keyname, char *keyval, const char *comm, char *card, int *status);
 int ffgnky(fitsfile *fptr, char *card, int *status);
 void ffcfmt(char *tform, char *cform);
 void ffcdsp(char *tform, char *cform);
@@ -312,8 +344,7 @@ int ffpbytoff(fitsfile *fptr, long gsize, long ngroups, long offset,
 int ffldrc(fitsfile *fptr, long record, int err_mode, int *status);
 int ffwhbf(fitsfile *fptr, int *nbuff);
 int ffbfeof(fitsfile *fptr, int *status);
-int ffbfwt(int nbuff, int *status);
-int fits_get_num_files(void);
+int ffbfwt(FITSfile *Fptr, int nbuff, int *status);
 int ffpxsz(int datatype);
 
 int ffourl(char *url, char *urltype, char *outfile, char *tmplfile,
@@ -878,6 +909,9 @@ int  ffffrw_work( long totalrows, long offset, long firstrow,
                   long nrows, int nCols, iteratorCol *colData,
                   void *userPtr );
 
+int fits_translate_pixkeyword(char *inrec, char *outrec,char *patterns[][2],
+    int npat, int naxis, int *colnum, int *pat_num, int *i,
+      int *j, int *n, int *m, int *l, int *status);
 
 /*  image compression routines */
 int fits_write_compressed_img(fitsfile *fptr, 
@@ -952,11 +986,11 @@ int imcomp_merge_overlap (char *tile, int pixlen, int ndim,
          long *fpixel, long *lpixel, int nullcheck, int *status);
 int imcomp_decompress_img(fitsfile *infptr, fitsfile *outfptr, int datatype,
          int  *status);
-int fits_quantize_float (float fdata[], long nx, long ny, int nullcheck,
+int fits_quantize_float (long row, float fdata[], long nx, long ny, int nullcheck,
          float in_null_value,
            float quantize_level, int idata[], double *bscale, double *bzero,
            int *iminval, int *imaxval);
-int fits_quantize_double (double fdata[], long nx, long ny, int nullcheck,
+int fits_quantize_double (long row, double fdata[], long nx, long ny, int nullcheck,
          double in_null_value,
            float quantize_level, int idata[], double *bscale, double *bzero,
            int *iminval, int *imaxval);
@@ -971,11 +1005,13 @@ int fits_rdecomp_byte (unsigned char *c, int clen, unsigned char array[], int nx
              int nblock);
 int pl_p2li (int *pxsrc, int xs, short *lldst, int npix);
 int pl_l2pi (short *ll_src, int xs, int *px_dst, int npix);
+int fits_init_randoms(void);
+
+int fitsio_init_lock(void);
 
 /* general driver routines */
 
 int urltype2driver(char *urltype, int *driver);
-int fits_init_cfitsio(void);
 
 int fits_register_driver( char *prefix,
 	int (*init)(void),
@@ -1017,6 +1053,17 @@ int file_read (int driverhandle, void *buffer, long nbytes);
 int file_write(int driverhandle, void *buffer, long nbytes);
 int file_is_compressed(char *filename);
 
+/* stream driver I/O routines */
+
+int stream_open(char *filename, int rwmode, int *driverhandle);
+int stream_create(char *filename, int *driverhandle);
+int stream_size(int driverhandle, LONGLONG *filesize);
+int stream_close(int driverhandle);
+int stream_flush(int driverhandle);
+int stream_seek(int driverhandle, LONGLONG offset);
+int stream_read (int driverhandle, void *buffer, long nbytes);
+int stream_write(int driverhandle, void *buffer, long nbytes);
+
 /* memory driver I/O routines */
 
 int mem_init(void);
