@@ -1,19 +1,18 @@
 /* FPACK utility routines
- * R. Seaman, NOAO & W. Pence, NASA/GSFC
- */
+   R. Seaman, NOAO & W. Pence, NASA/GSFC
+*/
 
 #include <time.h>
 #include <float.h>
 #include <signal.h>
+#include <sys/time.h>
+#include <math.h>
 #include "fitsio.h"
 #include "fpack.h"
-#include <sys/time.h>
-#include <time.h>
-#include <stdlib.h>
 
 /* these filename buffer are used to delete temporary files */
 /* in case the program is aborted */
-char tempfilename[SZ_STR];    
+char tempfilename[SZ_STR];
 char tempfilename2[SZ_STR];
 char tempfilename3[SZ_STR];
 
@@ -41,31 +40,88 @@ FILE *outreport;
 /* dimension of central image area to be sampled for test statistics */
 int XSAMPLE = 4100;
 int YSAMPLE = 4100;
-    
+
 int fp_msg (char *msg)
 {
-	printf ("%s", msg);
-	return(0);
+        printf ("%s", msg);
+        return(0);
 }
 /*--------------------------------------------------------------------------*/
 int fp_noop (void)
-{ 
-	fp_msg ("Input and output files are unchanged.\n");
-	return(0);
+{
+        fp_msg ("Input and output files are unchanged.\n");
+        return(0);
 }
 /*--------------------------------------------------------------------------*/
-int fp_version (void) 
-{ 
+int fp_version (void)
+{
         float version;
-	char cfitsioversion[20];
+        char cfitsioversion[40];
 
-	fp_msg (FPACK_VERSION);
-	fits_get_version(&version);
-	sprintf(cfitsioversion, " CFITSIO version %6.3f", version);
-	fp_msg(cfitsioversion);
-	fp_msg ("\n");
+        fp_msg (FPACK_VERSION);
+        fits_get_version(&version);
+        sprintf(cfitsioversion, " CFITSIO version %5.3f", version);
+        fp_msg(cfitsioversion);
+        fp_msg ("\n");
 }
+/*--------------------------------------------------------------------------*/
+int fp_access (char *filename)
+{
+	/* test if a file exists */
+	
+	FILE *diskfile;
+	
+	diskfile = fopen(filename, "r");
 
+	if (diskfile) {
+		fclose(diskfile);
+		return(0);
+	} else {
+	        return(-1);
+	}
+}
+/*--------------------------------------------------------------------------*/
+int fp_tmpnam(char *prefix, char *rootname, char *tmpnam)
+{
+	/* create temporary file name */
+
+	int maxtry = 30, len, i1 = 0, ii;
+
+	strcpy(tmpnam, prefix);  /*start with the prefix */
+	
+	len = strlen(rootname);
+	if (len > 64)
+	    i1 += (len - 64); 
+
+	/* append up to the last 64 characters of the rootname to the tmpnam */
+	strcat(tmpnam, rootname+i1);
+	
+	len = strlen(tmpnam);
+
+	/* replace all non-alphanumeric characters with 'x' (e.g. '/' or ':') */
+	for (ii = 0; ii < len; ii++) {
+	    if (!isalnum(tmpnam[ii]) )
+		tmpnam[ii] = 'x';
+	}
+
+        for (ii = 0; ii < maxtry; ii++) {
+		if (fp_access(tmpnam)) break;  /* good, the file does not exist */
+		strcat(tmpnam, "x");  /* append an x to the name, and try again */
+	}
+
+	if (ii == maxtry) {
+		len = strlen(tmpnam);
+		if (len > 75)
+		    tmpnam[75] = 0;  /* limit the length of the output string */
+
+		fp_msg ("\nCould not create temporary file name:\n");
+		fp_msg (tmpnam);
+		fp_msg ("\n");
+		exit (-1);
+	}
+
+        return(0);
+}
 /*--------------------------------------------------------------------------*/
 int fp_init (fpstate *fpptr)
 {
@@ -73,6 +129,8 @@ int fp_init (fpstate *fpptr)
 
 	fpptr->comptype = RICE_1;
 	fpptr->quantize_level = DEF_QLEVEL;
+        fpptr->no_dither = 0;
+        fpptr->dither_offset = 0;
 	fpptr->scale = DEF_HCOMP_SCALE;
 	fpptr->smooth = DEF_HCOMP_SMOOTH;
 	fpptr->rescale_noise = DEF_RESCALE_NOISE;
@@ -92,6 +150,7 @@ int fp_init (fpstate *fpptr)
 	fpptr->verbose = 0;
 
 	fpptr->prefix[0] = (char) NULL;
+	fpptr->extname[0] = (char) NULL;
 	fpptr->delete_suffix = 0;
 	fpptr->outfile[0] = (char) NULL;
 
@@ -103,13 +162,13 @@ int fp_init (fpstate *fpptr)
 	fpptr->preflight_checked = 0;
 	return(0);
 }
-
 /*--------------------------------------------------------------------------*/
 int fp_list (int argc, char *argv[], fpstate fpvar)
 {
 	fitsfile *infptr;
-	char	infits[SZ_STR];
+	char	infits[SZ_STR], msg[SZ_STR];
 	int	hdunum, iarg, stat=0;
+	LONGLONG sizell;
 
 	if (fpvar.initialized != FP_INIT_MAGIC) {
 	    fp_msg ("Error: internal initialization error\n"); exit (-1);
@@ -123,19 +182,30 @@ int fp_list (int argc, char *argv[], fpstate fpvar)
 		fp_msg (infits); fp_msg ("\n"); exit (-1);
 	    }
 
-	    if (access (infits, R_OK) != 0) {
-		fp_msg ("Error: can't find or open input file ");
-		fp_msg (infits); fp_msg ("\n"); exit (-1);
+	    if (fp_access (infits) != 0) {
+		        fp_msg ("Error: can't find or read input file "); fp_msg (infits);
+		        fp_msg ("\n"); fp_noop (); exit (-1);
 	    }
 
 	    fits_open_file (&infptr, infits, READONLY, &stat);
 	    if (stat) { fits_report_error (stderr, stat); exit (stat); }
 
-	    fp_info (infits);
-
+	    /* move to the end of file, to get the total size in bytes */
 	    fits_get_num_hdus (infptr, &hdunum, &stat);
+	    fits_movabs_hdu (infptr, hdunum, NULL, &stat);
+	    fits_get_hduaddrll(infptr, NULL, NULL, &sizell, &stat);
 	    if (stat) { fits_report_error (stderr, stat); exit (stat); }
 
+            sprintf (msg, "# %s (", infits); fp_msg (msg);
+
+#if defined(_MSC_VER)
+    /* Microsoft Visual C++ 6.0 uses '%I64d' syntax  for 8-byte integers */
+        sprintf(msg, "%I64d bytes)\n", sizell); fp_msg (msg);
+#elif (USE_LL_SUFFIX == 1)
+        sprintf(msg, "%lld bytes)\n", sizell); fp_msg (msg);
+#else
+        sprintf(msg, "%ld bytes)\n", sizell); fp_msg (msg);
+#endif
 	    fp_info_hdu (infptr);
 
 	    fits_close_file (infptr, &stat);
@@ -143,56 +213,34 @@ int fp_list (int argc, char *argv[], fpstate fpvar)
 	}
 	return(0);
 }
-
-/*--------------------------------------------------------------------------*/
-int fp_info (char *infits)
-{
-	struct  stat    sbuf;
-	char	msg[SZ_STR];
-	int     mtime, size, uid, gid, nlink;
-	unsigned mode;
-
-	if (stat (infits, &sbuf) != 0) {
-	    fp_msg ("Error: can't stat "); fp_msg (infits); fp_msg ("\n");
-
-	} else {
-            size = (int) sbuf.st_size;
-            mtime = (int) sbuf.st_mtime;
-            uid = (int) sbuf.st_uid;
-            gid = (int) sbuf.st_gid;
-            mode = (unsigned) sbuf.st_mode;
-            nlink = (int) sbuf.st_nlink;
-
-	    sprintf (msg, "%s: %d bytes\n", infits, size); fp_msg (msg);
-	}
-	return(0);
-}
-
 /*--------------------------------------------------------------------------*/
 int fp_info_hdu (fitsfile *infptr)
 {
 	long	naxes[9] = {1, 1, 1, 1, 1, 1, 1, 1, 1};
 	char	msg[SZ_STR], val[SZ_CARD], com[SZ_CARD];
-	int	naxis=0, hdutype, bitpix, hdupos, stat=0, ii;
+        int     comptype, naxis=0, hdutype, bitpix, hdupos, stat=0, ii;
+        unsigned long   datasum, hdusum;
 
 	fits_movabs_hdu (infptr, 1, NULL, &stat);
 	if (stat) { fits_report_error (stderr, stat); exit (stat); }
 
 	for (hdupos=1; ! stat; hdupos++) {
 	    fits_get_hdu_type (infptr, &hdutype, &stat);
-/*	    fits_read_keyword (infptr, FILE_KEY, val, com, &stat);
- */
+
 	    if (stat) { fits_report_error (stderr, stat); exit (stat); }
 
+            fits_get_chksum(infptr, &datasum, &hdusum, &stat);
+
 	    if (hdutype == IMAGE_HDU) {
 		sprintf (msg, "  %d IMAGE", hdupos); fp_msg (msg);
+                sprintf (msg, " SUMS=%u/%u", ~hdusum, datasum); fp_msg (msg);
 
-/*		sprintf (msg, "  %d IMAGE %s", hdupos, val); fp_msg (msg);
- */
 		fits_get_img_param (infptr, 9, &bitpix, &naxis, naxes, &stat);
 
+                sprintf (msg, " BITPIX=%d", bitpix); fp_msg (msg);
+
 		if (naxis == 0) {
-		    sprintf (msg, " [no pixels]"); fp_msg (msg);
+                    sprintf (msg, " [no_pixels]"); fp_msg (msg);
 		} else if (naxis == 1) {
 		    sprintf (msg, " [%d]", naxes[1]); fp_msg (msg);
 		} else {
@@ -203,20 +251,36 @@ int fp_info_hdu (fitsfile *infptr)
 		    fp_msg ("]");
 		}
 
-		if (fits_is_compressed_image (infptr, &stat))
-		    fp_msg (" (compressed)\n");
-		else
-		    fp_msg ("\n");
+                if (fits_is_compressed_image (infptr, &stat)) {
+                    fits_read_keyword (infptr, "ZCMPTYPE", val, com, &stat);
 
-	    } else if (hdutype == ASCII_TBL) {
-		sprintf (msg, "  %d ASCII TABLE\n", hdupos); fp_msg (msg);
+                    /* allow for quote in keyword value */
+                    if (! strncmp (val+1, "RICE_1", 6))
+                        fp_msg (" tiled_rice\n");
+                    else if (! strncmp (val+1, "GZIP_1", 6))
+                        fp_msg (" tiled_gzip\n");
+                    else if (! strncmp (val+1, "PLIO_1", 6))
+                        fp_msg (" tiled_plio\n");
+                    else if (! strncmp (val+1, "HCOMPRESS_1", 11))
+                        fp_msg (" tiled_hcompress\n");
+                    else
+                        fp_msg (" unknown\n");
 
-	    } else if (hdutype == BINARY_TBL) {
-		sprintf (msg, "  %d BINARY TABLE\n", hdupos); fp_msg (msg);
+                } else
+                    fp_msg (" not_tiled\n");
 
-	    } else {
-		sprintf (msg, "  %d UNKNOWN EXTENSION\n", hdupos); fp_msg (msg);
-	    }
+            } else if (hdutype == ASCII_TBL) {
+                sprintf (msg, "  %d ASCII_TBL\n", hdupos); fp_msg (msg);
+                sprintf (msg, " SUMS=%u/%u", ~hdusum, datasum); fp_msg (msg);
+
+            } else if (hdutype == BINARY_TBL) {
+                sprintf (msg, "  %d BINARY_TBL\n", hdupos); fp_msg (msg);
+                sprintf (msg, " SUMS=%u/%u", ~hdusum, datasum); fp_msg (msg);
+
+            } else {
+                sprintf (msg, "  %d OTHER\n", hdupos); fp_msg (msg);
+                sprintf (msg, " SUMS=%u/%u", ~hdusum, datasum); fp_msg (msg);
+            }
 
 	    fits_movrel_hdu (infptr, 1, NULL, &stat);
 	}
@@ -253,18 +317,19 @@ int fp_preflight (int argc, char *argv[], int unpack, fpstate *fpptr)
 	  	/* ********** This section applies to funpack ************ */
 
 	      /* check that input file  exists */
-	      if (access (infits, R_OK) != 0) {  /* if not, then check if */
+	      if (infits[0] != '-') {  /* if not reading from stdin stream */
+	         if (fp_access (infits) != 0) {  /* if not, then check if */
 		    strcat(infits, ".fz");       /* a .fz version exsits */
-	            if (access (infits, R_OK) != 0) {
+	            if (fp_access (infits) != 0) {
                         namelen = strlen(infits);
                         infits[namelen - 3] = '\0';  /* remove the .fz suffix */
 		        fp_msg ("Error: can't find or read input file "); fp_msg (infits);
 		        fp_msg ("\n"); fp_noop (); exit (-1);
                     }
-	      } else {   /* make sure a .fz version of the same file doesn't exist */
+	         } else {   /* make sure a .fz version of the same file doesn't exist */
                     namelen = strlen(infits);
 		    strcat(infits, ".fz");   
-	            if (access (infits, R_OK) == 0) {
+	            if (fp_access (infits) == 0) {
                         infits[namelen] = '\0';  /* remove the .fz suffix */
 		        fp_msg ("Error: ambiguous input file name.  Which file should be unpacked?:\n  ");
 		        fp_msg (infits); fp_msg ("\n  "); 
@@ -273,6 +338,7 @@ int fp_preflight (int argc, char *argv[], int unpack, fpstate *fpptr)
                     } else {
                         infits[namelen] = '\0';  /* remove the .fz suffix */
 		    }
+	         }
 	      }
 
               /* if writing to stdout, then we are all done */
@@ -289,7 +355,7 @@ int fp_preflight (int argc, char *argv[], int unpack, fpstate *fpptr)
 	          }
 
                   /* check that output file doesn't exist */
-	          if (access (fpptr->outfile, F_OK) == 0) {
+	          if (fp_access (fpptr->outfile) == 0) {
 		            fp_msg ("Error: output file already exists:\n "); 
 			    fp_msg (fpptr->outfile);
 		            fp_msg ("\n "); fp_noop (); exit (-1);
@@ -306,7 +372,12 @@ int fp_preflight (int argc, char *argv[], int unpack, fpstate *fpptr)
 	          strcat(outfits,fpptr->prefix);
 	      }
 
-	      strcat(outfits, infits);
+	      /* construct output file name */
+	      if (infits[0] == '-') {
+	        strcpy(outfits, "output.fits");
+	      } else {
+	        strcpy(outfits, infits);
+	      }
 
 	      /* remove .gz suffix, if present (output is not gzipped) */
               namelen = strlen(outfits);
@@ -316,18 +387,20 @@ int fp_preflight (int argc, char *argv[], int unpack, fpstate *fpptr)
 
 	      /* check for .fz suffix that is sometimes required */
 	      /* and remove it if present */
-              namelen = strlen(outfits);
-	      if ( !strcmp(".fz", outfits + namelen - 3) ) { /* suffix is present */
+	      if (infits[0] != '-') {  /* if not reading from stdin stream */
+                 namelen = strlen(outfits);
+	         if ( !strcmp(".fz", outfits + namelen - 3) ) { /* suffix is present */
                         outfits[namelen - 3] = '\0';
-	      } else if (fpptr->delete_suffix) {  /* required suffix is missing */
+	         } else if (fpptr->delete_suffix) {  /* required suffix is missing */
 		    fp_msg ("Error: input compressed file "); fp_msg (infits);
 		    fp_msg ("\n does not have the default .fz suffix.\n"); 
 		    fp_noop (); exit (-1);
+	         }
 	      }
 
 	      /* if infits != outfits, make sure outfits doesn't already exist */
               if (strcmp(infits, outfits)) {
-	                if (access (outfits, F_OK) == 0) {
+	                if (fp_access (outfits) == 0) {
 		            fp_msg ("Error: output file already exists:\n "); fp_msg (outfits);
 		            fp_msg ("\n "); fp_noop (); exit (-1);
 		        }       
@@ -336,7 +409,7 @@ int fp_preflight (int argc, char *argv[], int unpack, fpstate *fpptr)
 	      /* if gzipping the output, make sure .gz file doesn't exist */
 	      if (fpptr->do_gzip_file) {
 	                strcat(outfits, ".gz");
-	                if (access (outfits, F_OK) == 0) {
+	                if (fp_access (outfits) == 0) {
 		            fp_msg ("Error: output file already exists:\n "); fp_msg (outfits);
 		            fp_msg ("\n "); fp_noop (); exit (-1);
 		        }       
@@ -347,14 +420,16 @@ int fp_preflight (int argc, char *argv[], int unpack, fpstate *fpptr)
 	  	/* ********** This section applies to fpack ************ */
 
 	      /* check that input file  exists */
-	      if (access (infits, R_OK) != 0) {  /* if not, then check if */
+	      if (infits[0] != '-') {  /* if not reading from stdin stream */
+	        if (fp_access (infits) != 0) {  /* if not, then check if */
 		    strcat(infits, ".gz");     /* a gzipped version exsits */
-	            if (access (infits, R_OK) != 0) {
+	            if (fp_access (infits) != 0) {
                         namelen = strlen(infits);
                         infits[namelen - 3] = '\0';  /* remove the .gz suffix */
 		        fp_msg ("Error: can't find or read input file "); fp_msg (infits);
 		        fp_msg ("\n"); fp_noop (); exit (-1);
                     }
+	        }
 	      }
 
               /* make sure the file to pack does not already have a .fz suffix */
@@ -370,21 +445,32 @@ int fp_preflight (int argc, char *argv[], int unpack, fpstate *fpptr)
 	      }
 
 	      /* construct output file name */
-	      strcpy(outfits, infits);
-	      
+	      if (infits[0] == '-') {
+	        strcpy(outfits, "input.fits");
+	      } else {
+	        strcpy(outfits, infits);
+	      }
+
 	      /* remove .gz suffix, if present (output is not gzipped) */
               namelen = strlen(outfits);
 	      if ( !strcmp(".gz", outfits + namelen - 3) ) {
                         outfits[namelen - 3] = '\0';
 	      }
 	      
+	      /* remove .imh suffix (IRAF format image), and replace with .fits */
+              namelen = strlen(outfits);
+	      if ( !strcmp(".imh", outfits + namelen - 4) ) {
+                        outfits[namelen - 4] = '\0';
+                        strcat(outfits, ".fits");
+	      }
+
 	      /* If not clobbering the input file, add .fz suffix to output name */
 	      if (! fpptr->clobber)
 		        strcat(outfits, ".fz");
 			
 	      /* if infits != outfits, make sure outfits doesn't already exist */
               if (strcmp(infits, outfits)) {
-	                if (access (outfits, F_OK) == 0) {
+	                if (fp_access (outfits) == 0) {
 		            fp_msg ("Error: output file already exists:\n "); fp_msg (outfits);
 		            fp_msg ("\n "); fp_noop (); exit (-1);
 		        }       
@@ -403,7 +489,8 @@ int fp_loop (int argc, char *argv[], int unpack, fpstate fpvar)
 {
 	char	infits[SZ_STR], outfits[SZ_STR];
 	char	temp[SZ_STR], answer[30], *cptr;
-	int	iarg, islossless, namelen;
+	int	ii, iarg, islossless, namelen, iraf_infile = 0, status = 0;
+	FILE	*diskfile;
         
 	if (fpvar.initialized != FP_INIT_MAGIC) {
 	    fp_msg ("Error: internal initialization error\n"); exit (-1);
@@ -411,11 +498,10 @@ int fp_loop (int argc, char *argv[], int unpack, fpstate fpvar)
 	    fp_msg ("Error: internal preflight error\n"); exit (-1);
 	}
 
-
 	if (fpvar.test_all && fpvar.outfile[0]) {
 	    outreport = fopen(fpvar.outfile, "w");
 		fprintf(outreport," Filename Extension BITPIX NAXIS1 NAXIS2 Size N_nulls Minval Maxval Mean Sigm Noise1 Noise3 T_whole T_rowbyrow ");
-		fprintf(outreport,"[Comp_ratio, Pack_cpu, Unpack_cpu, Lossless readtimes] (repeated for Rice, Hcompress and GZIP)\n");
+		fprintf(outreport,"[Comp_ratio, Pack_cpu, Unpack_cpu, Lossless readtimes] (repeated for Rice, Hcompress, and GZIP)\n");
 	}
 
 
@@ -454,8 +540,10 @@ int fp_loop (int argc, char *argv[], int unpack, fpstate fpvar)
 	  	/* ********** This section applies to funpack ************ */
 
 	      /* find input file */
-	      if (access (infits, R_OK) != 0) {  /* if not, then */
+	      if (infits[0] != '-') {  /* if not reading from stdin stream */
+	         if (fp_access (infits) != 0) {  /* if not, then */
 		    strcat(infits, ".fz");       /* a .fz version must exsit */
+	         }
 	      }
 
 	      if (fpvar.to_stdout) {
@@ -470,7 +558,12 @@ int fp_loop (int argc, char *argv[], int unpack, fpstate fpvar)
 	              strcat(outfits,fpvar.prefix);
 	          }
 
-	          strcat(outfits, infits);
+	          /* construct output file name */
+	          if (infits[0] == '-') {
+	            strcpy(outfits, "output.fits");
+	          } else {
+	            strcpy(outfits, infits);
+	          }
 
 	          /* remove .gz suffix, if present (output is not gzipped) */
                   namelen = strlen(outfits);
@@ -492,8 +585,13 @@ int fp_loop (int argc, char *argv[], int unpack, fpstate fpvar)
 	      if (fpvar.to_stdout) {
 		strcpy(outfits, "-");
 	      } else if (! fpvar.test_all) {
+
 	          /* construct output file name */
-	          strcpy(outfits, infits);
+	          if (infits[0] == '-') {
+	            strcpy(outfits, "input.fits");
+	          } else {
+	            strcpy(outfits, infits);
+	          }
 	      
 	          /* remove .gz suffix, if present (output is not gzipped) */
                   namelen = strlen(outfits);
@@ -501,6 +599,15 @@ int fp_loop (int argc, char *argv[], int unpack, fpstate fpvar)
                         outfits[namelen - 3] = '\0';
 	          }
 	      
+	          /* remove .imh suffix (IRAF format image), and replace with .fits */
+                  namelen = strlen(outfits);
+	          if ( !strcmp(".imh", outfits + namelen - 4) ) {
+                        outfits[namelen - 4] = '\0';
+                        strcat(outfits, ".fits");
+                        iraf_infile = 1;  /* this is an IRAF format input file */
+			           /* change the output name to "NAME.fits.fz" */
+	          }
+
 	          /* If not clobbering the input file, add .fz suffix to output name */
 	          if (! fpvar.clobber)
 		        strcat(outfits, ".fz");
@@ -509,7 +616,8 @@ int fp_loop (int argc, char *argv[], int unpack, fpstate fpvar)
 
           strncpy(temp, outfits, SZ_STR-1);
 
-          if (!strcmp(infits, outfits) ) {  /* are input and output names the same? */
+	  if (infits[0] != '-') {  /* if not reading from stdin stream */
+             if (!strcmp(infits, outfits) ) {  /* are input and output names the same? */
 
                 /* clobber the input file with the output file with the same name */
                 if (! fpvar.clobber) {
@@ -517,19 +625,11 @@ int fp_loop (int argc, char *argv[], int unpack, fpstate fpvar)
 		    exit (-1);
 		} 
 
-	        /* create temporary output file */
-		strcpy (outfits, "fpack_tmp.XXXXXX"); mktemp (outfits);
-
-		if (access (outfits, F_OK) == 0) {
-		    /* unlikely name collision, try again (once) */
-		    strcpy (outfits, "fpack_tmp.XXXXXX"); mktemp (outfits);
-
-		    if (access (outfits, F_OK) == 0) {
-			fp_msg ("Error: temporary file "); fp_msg (outfits);
-			fp_msg (" already exists\n"); exit (-1);
-		    }
-		}
+                /* create temporary file name */
+		fp_tmpnam("fpTmpfile1", infits, outfits);
+		
                 strcpy(tempfilename, outfits);  /* store temp file name, in case of abort */
+	      }
 	    }
 
 
@@ -540,8 +640,9 @@ int fp_loop (int argc, char *argv[], int unpack, fpstate fpvar)
 		
 	    if (fpvar.test_all) {   /* compare all the algorithms */
 
-		strcpy (tempfilename,  "fpack_tmp.XXXXXX"); mktemp (tempfilename);
-		strcpy (tempfilename2, "fpack_tmp.XXXXXX"); mktemp (tempfilename2);
+                /* create 2 temporary file names */
+		fp_tmpnam("fpTmpfile1", infits, tempfilename);
+		fp_tmpnam("fpTmpfile2", infits, tempfilename2);
 
 		fp_test (infits, tempfilename, tempfilename2, fpvar);
 
@@ -552,7 +653,17 @@ int fp_loop (int argc, char *argv[], int unpack, fpstate fpvar)
                 continue;
 
 	    } else if (unpack) {
-		fp_unpack (infits, outfits, fpvar);
+		/* unpack to temporary file, so other tasks can't open it until it is renamed */
+
+                /* create  temporary file name */
+		fp_tmpnam("fpTmpfile2", infits, tempfilename2);
+
+		/* unpack the input file to the temporary file */
+		fp_unpack (infits, tempfilename2, fpvar);
+
+		/* rename the temporary file to it's real name */
+		rename(tempfilename2, outfits);
+		tempfilename2[0] = '\0';  /* clear temporary file name */
 
 	    }  else {
 		fp_pack (infits, outfits, fpvar, &islossless);
@@ -580,6 +691,14 @@ int fp_loop (int argc, char *argv[], int unpack, fpstate fpvar)
 		}
  
 	        /* rename clobbers input, may be unix/shell version dependent */
+
+		if (iraf_infile) {  /* special case of deleting an IRAF format header and pixel file */
+		   if (fits_delete_iraf_file(infits, &status)) {
+		        fp_msg("\nError deleting IRAF .imh and .pix files.\n");
+			fp_msg(infits); fp_msg ("\n"); exit (-1);
+		    }
+		}
+				
 		if (rename (outfits, temp) != 0) {
 		        fp_msg ("\nError renaming tmp file to ");
 		        fp_msg (temp); fp_msg ("\n"); exit (-1);
@@ -588,25 +707,38 @@ int fp_loop (int argc, char *argv[], int unpack, fpstate fpvar)
                 strcpy(outfits, temp);
 
 	    } else if (fpvar.clobber || fpvar.delete_input) {      /* delete the input file */
-	         if (!islossless && !fpvar.do_not_prompt) {
+	         if (!islossless && !fpvar.do_not_prompt) {  /* user did not turn off delete prompt */
 		    fp_msg ("\nFile ");
 		    fp_msg (infits); 
 		    fp_msg ("\nwas compressed with a LOSSY method.  \n");
 		    fp_msg ("Delete the original file? (Y/N) ");
 		    fgets(answer, 29, stdin);
-		    if (answer[0] != 'Y' && answer[0] != 'y') {
+		    if (answer[0] != 'Y' && answer[0] != 'y') {  /* user abort */
 		        fp_msg ("\noriginal file NOT deleted!\n");
-		    } else if (remove(infits) != 0) {
-		        fp_msg ("\nError deleting input file ");
-		        fp_msg (infits); fp_msg ("\n"); exit (-1);
+		    } else {
+			if (iraf_infile) {  /* special case of deleting an IRAF format header and pixel file */
+		   	    if (fits_delete_iraf_file(infits, &status)) {
+		        	fp_msg("\nError deleting IRAF .imh and .pix files.\n");
+				fp_msg(infits); fp_msg ("\n"); exit (-1);
+			    }
+		        }  else if (remove(infits) != 0) {  /* normal case of deleting input FITS file */
+		            fp_msg ("\nError deleting input file ");
+		            fp_msg (infits); fp_msg ("\n"); exit (-1);
+		        }
 		    }
-		  } else {
-		     if (remove(infits) != 0) {
-		        fp_msg ("\nError deleting input file ");
-		        fp_msg (infits); fp_msg ("\n"); exit (-1);
-		     }
+		  } else {   /* user said don't prompt, so just delete the input file */
+			if (iraf_infile) {  /* special case of deleting an IRAF format header and pixel file */
+		   	    if (fits_delete_iraf_file(infits, &status)) {
+		        	fp_msg("\nError deleting IRAF .imh and .pix files.\n");
+				fp_msg(infits); fp_msg ("\n"); exit (-1);
+			    }
+		        }  else if (remove(infits) != 0) {  /* normal case of deleting input FITS file */
+		            fp_msg ("\nError deleting input file ");
+		            fp_msg (infits); fp_msg ("\n"); exit (-1);
+		        }
 		  }
 	    }
+            iraf_infile = 0; 
 
 	    if (fpvar.do_gzip_file) {       /* gzip the output file */
 		strcpy(temp, "gzip -1 ");
@@ -639,10 +771,15 @@ int fp_pack (char *infits, char *outfits, fpstate fpvar, int *islossless)
 	if (stat) { fits_report_error (stderr, stat); exit (stat); }
 
 	fits_set_compression_type (outfptr, fpvar.comptype, &stat);
+
+	if (fpvar.no_dither)
+	    fits_set_quantize_dither(outfptr, -1, &stat);
+	    
 	fits_set_quantize_level (outfptr, fpvar.quantize_level, &stat);
 	fits_set_hcomp_scale (outfptr, fpvar.scale, &stat);
 	fits_set_hcomp_smooth (outfptr, fpvar.smooth, &stat);
 	fits_set_tile_dim (outfptr, 6, fpvar.ntile, &stat);
+	fits_set_dither_offset(outfptr, fpvar.dither_offset, &stat);
 
 	if (stat) { fits_report_error (stderr, stat); exit (stat); }
 	
@@ -679,21 +816,121 @@ int fp_pack (char *infits, char *outfits, fpstate fpvar, int *islossless)
 int fp_unpack (char *infits, char *outfits, fpstate fpvar)
 {
         fitsfile *infptr, *outfptr;
-        int     stat=0;
+        int stat=0, hdutype, extnum, single = 0;
+	char *loc, *hduloc, hduname[SZ_STR];
 
         fits_open_file (&infptr, infits, READONLY, &stat);
         fits_create_file (&outfptr, outfits, &stat);
 
-        if (stat) { fits_report_error (stderr, stat); exit (stat); }
+        if (fpvar.extname[0]) {  /* unpack a list of HDUs? */
+
+            /* move to the first HDU in the list */
+	    hduloc = fpvar.extname;
+	    loc = strchr(hduloc, ','); /* look for 'comma' delimiter between names */
+	    
+	    if (loc)         
+	        *loc = '\0';  /* terminate the first name in the string */
+
+	    strcpy(hduname, hduloc);  /* copy the first name into temporary string */
+
+	    if (loc)        
+	        hduloc = loc + 1;  /* advance to the beginning of the next name, if any */
+            else {
+	        hduloc += strlen(hduname);  /* end of the list */
+	        single = 1;  /* only 1 HDU is being unpacked */
+            }
+
+            if (isdigit( (int) hduname[0]) ) {
+	       extnum = strtol(hduname, &loc, 10); /* read the string as an integer */
+
+               /* check for junk following the integer */
+               if (*loc == '\0' )  /* no junk, so move to this HDU number (+1) */
+               {	       
+                  fits_movabs_hdu(infptr, extnum + 1, &hdutype, &stat);  /* move to HDU number */
+                  if (hdutype != IMAGE_HDU)
+		     stat = NOT_IMAGE;
+
+               } else {  /* the string is not an integer, so must be the column name */
+	          hdutype = IMAGE_HDU;
+	          fits_movnam_hdu(infptr, hdutype, hduname, 0, &stat);
+ 	       }
+            }
+	    else
+	    {
+                /* move to the named image extension */
+  	        hdutype = IMAGE_HDU;
+	        fits_movnam_hdu(infptr, hdutype, hduname, 0, &stat);
+            }
+        }
+        if (stat) {
+	  fp_msg ("Unable to find and move to extension '");
+          fp_msg(hduname);
+	  fp_msg("'\n");
+	  fits_report_error (stderr, stat); 
+	  exit (stat);
+        }
 
         while (! stat) {
+	
+	    if (single)
+	        stat = -1;  /* special status flag to force output primary array */
+
             fp_unpack_hdu (infptr, outfptr, &stat);
 
 	    if (fpvar.do_checksums) {
 	        fits_write_chksum (outfptr, &stat);
 	    }
 
-            fits_movrel_hdu (infptr, 1, NULL, &stat);
+	    /* move to the next HDU */
+            if (fpvar.extname[0]) {  /* unpack a list of HDUs? */
+
+                if (!(*hduloc)) {
+		    stat = END_OF_FILE;  /* we reached the end of the list */
+		} else {
+		    /* parse the next HDU name and move to it */
+	            loc = strchr(hduloc, ',');
+	    
+	            if (loc)         /* look for 'comma' delimiter between names */
+	               *loc = '\0';  /* terminate the first name in the string */
+
+	            strcpy(hduname, hduloc);  /* copy the next name into temporary string */
+
+	            if (loc)         
+	                hduloc = loc + 1;  /* advance to the beginning of the next name, if any */
+                    else
+	               *hduloc = '\0';  /* end of the list */
+
+                    if (isdigit( (int) hduname[0]) ) {
+	               extnum = strtol(hduname, &loc, 10); /* read the string as an integer */
+
+                      /* check for junk following the integer */
+                      if (*loc == '\0' )   /* no junk, so move to this HDU number (+1) */
+		      {	       
+                        fits_movabs_hdu(infptr, extnum + 1, &hdutype, &stat);  /* move to HDU number */
+                        if (hdutype != IMAGE_HDU)
+		        stat = NOT_IMAGE;
+
+                      } else {  /* the string is not an integer, so must be the column name */
+	                hdutype = IMAGE_HDU;
+	                fits_movnam_hdu(infptr, hdutype, hduname, 0, &stat);
+ 	              }
+
+                    } else {
+                      /* move to the named image extension */
+  	              hdutype = IMAGE_HDU;
+	              fits_movnam_hdu(infptr, hdutype, hduname, 0, &stat);
+                   }
+
+                   if (stat) {
+	              fp_msg ("Unable to find and move to extension '");
+                      fp_msg(hduname);
+	              fp_msg("'\n");
+                   }
+                }
+            } else {
+                /* increment to the next HDU */
+                fits_movrel_hdu (infptr, 1, NULL, &stat);
+            }
         }
 
         if (stat == END_OF_FILE) stat = 0;
@@ -727,6 +964,7 @@ int fp_test (char *infits, char *outfits, char *outfits2, fpstate fpvar)
 	double  bscale, rescale;
 	long headstart, datastart, dataend;
 	float origdata = 0., whole_cpu, whole_elapse, row_elapse, row_cpu, xbits;
+	FILE	*diskfile;
 
 	fits_open_file (&inputfptr, infits, READONLY, &stat);
 	fits_create_file (&outfptr, outfits, &stat);
@@ -734,10 +972,14 @@ int fp_test (char *infits, char *outfits, char *outfits2, fpstate fpvar)
 
 	if (stat) { fits_report_error (stderr, stat); exit (stat); }
 
+	if (fpvar.no_dither)
+	    fits_set_quantize_dither(outfptr, -1, &stat);
+
 	fits_set_quantize_level (outfptr, fpvar.quantize_level, &stat);
 	fits_set_hcomp_scale (outfptr, fpvar.scale, &stat);
 	fits_set_hcomp_smooth (outfptr, fpvar.smooth, &stat);
 	fits_set_tile_dim (outfptr, 6, fpvar.ntile, &stat);
+	fits_set_dither_offset(outfptr, fpvar.dither_offset, &stat);
 
 	while (! stat) {
 	
@@ -773,7 +1015,9 @@ int fp_test (char *infits, char *outfits, char *outfits2, fpstate fpvar)
 			  /* all the criteria are met, so create a temporary file that */
 			  /* contains a rescaled version of the image */
 			  
-			  strcpy (tempfilename3, "fptmp.XXXXXX"); mktemp (tempfilename3);
+                	  /* create temporary file name */
+			  fp_tmpnam("fpTmpfile3", infits, tempfilename3);
+
 			  fits_create_file(&tempfile, tempfilename3, &stat);
 
 			  fits_get_hdu_num(inputfptr, &hdunum);
@@ -885,6 +1129,12 @@ int fp_test (char *infits, char *outfits, char *outfits2, fpstate fpvar)
 		fits_set_compression_type (outfptr, GZIP_1, &stat);
 		fits_set_tile_dim (outfptr, 6, fpvar.ntile, &stat);
 		fp_test_hdu(infptr, outfptr, outfptr2, fpvar, &stat);
+
+/*
+		fits_set_compression_type (outfptr, BZIP2_1, &stat);
+		fits_set_tile_dim (outfptr, 6, fpvar.ntile, &stat);
+		fp_test_hdu(infptr, outfptr, outfptr2, fpvar, &stat);
+*/
 /*
 		fits_set_compression_type (outfptr, PLIO_1, &stat);
 		fits_set_tile_dim (outfptr, 6, fpvar.ntile, &stat);
@@ -934,6 +1184,8 @@ int fp_pack_hdu (fitsfile *infptr, fitsfile *outfptr, fpstate fpvar,
 	int	stat=0, totpix=0, naxis=0, ii, hdutype, bitpix;
 	int	tstatus, hdunum, rescale_flag = 0;
 	double  bscale, rescale;
+	FILE	*diskfile;
+	char	infits[SZ_STR];
 
 	if (*status) return(0);
 
@@ -969,7 +1221,10 @@ int fp_pack_hdu (fitsfile *infptr, fitsfile *outfptr, fpstate fpvar,
 			  /* all the criteria are met, so create a temporary file that */
 			  /* contains a rescaled version of the image */
 			  
-			  strcpy (tempfilename3, "fptmp.XXXXXX"); mktemp (tempfilename3);
+			  /* create temporary file name */
+			  fits_file_name(infptr, infits, &stat);  /* get the input file name */
+			  fp_tmpnam("fpTmpfile3", infits, tempfilename3);
+
 			  fits_create_file(&tempfile, tempfilename3, &stat);
 
 			  fits_get_hdu_num(infptr, &hdunum);
@@ -1023,16 +1278,13 @@ int fp_pack_hdu (fitsfile *infptr, fitsfile *outfptr, fpstate fpvar,
 /*--------------------------------------------------------------------------*/
 int fp_unpack_hdu (fitsfile *infptr, fitsfile *outfptr, int *status)
 {
-        int stat=0;
+        if (*status > 0) return(0);
 
-        if (*status) return(0);
-
-        if (fits_is_compressed_image (infptr,  &stat))
-            fits_img_decompress (infptr, outfptr, &stat);
+        if (fits_is_compressed_image (infptr,  status))
+            fits_img_decompress (infptr, outfptr, status);
         else
-            fits_copy_hdu (infptr, outfptr, 0, &stat);
+            fits_copy_hdu (infptr, outfptr, 0, status);
 
-        *status = stat;
 	return(0);
 }
 /*--------------------------------------------------------------------------*/
@@ -1068,6 +1320,10 @@ int fits_read_image_speed (fitsfile *infptr, float *whole_elapse,
         if (bitpix == BYTE_IMG) {
 		carray = calloc(naxes[1]*naxes[0], sizeof(char));
 
+                /* remove any cached uncompressed tile 
+		  (dangerous to directly modify the structure!) */
+                (infptr->Fptr)->tilerow = 0;
+
 		marktime(status);
 		fits_read_subset(infptr, TBYTE, fpixel, lpixel, inc, &cnull, 
 		      carray, &anynull, status);
@@ -1077,6 +1333,11 @@ int fits_read_image_speed (fitsfile *infptr, float *whole_elapse,
 
 		/* now read the image again, row by row */
 		if (row_elapse) {
+
+                  /* remove any cached uncompressed tile 
+	  	    (dangerous to directly modify the structure!) */
+                  (infptr->Fptr)->tilerow = 0;
+
 		  marktime(status);
 		  for (ii = 0; ii < naxes[1]; ii++) {
 		   fpixel[1] = ii+1;
@@ -1222,12 +1483,22 @@ int fp_test_hdu (fitsfile *infptr, fitsfile *outfptr, fitsfile *outfptr2,
 		strcpy(ctype, "RICE");
 	else if (comptype == GZIP_1)
 		strcpy(ctype, "GZIP");
+/*
+	else if (comptype == BZIP2_1)
+		strcpy(ctype, "BZIP2");
+*/
 	else if (comptype == PLIO_1)
 		strcpy(ctype, "PLIO");
 	else if (comptype == HCOMPRESS_1)
 		strcpy(ctype, "HCOMP");
 	else if (comptype == NOCOMPRESS)
 		strcpy(ctype, "NONE");
+	else {
+	      fp_msg ("Error: unsupported image compression type ");
+	      *status = DATA_COMPRESSION_ERR;
+	      return(0);
+	}
+
 	/* -------------- COMPRESS the image ------------------ */
 
 	marktime(&stat);
@@ -1238,20 +1509,27 @@ int fp_test_hdu (fitsfile *infptr, fitsfile *outfptr, fitsfile *outfptr2,
 	gettime(&elapse, &packcpu, &stat);
 
 	/* get elapsed and cpu times need to read the compressed image */
-/*
-	if (comptype == HCOMPRESS_1) {
+
+        /* if whole image is compressed as single tile, don't read row by row
+	   because it usually takes a very long time
+	*/
+        if (fpvar.ntile[1] == 0) {
 	  fits_read_image_speed (outfptr, &whole_elapse, &whole_cpu, 
 	   0, 0, &stat);
 	  row_elapse = 0; row_cpu = 0;
-	} else 
-*/
+	} else {
+
 	  fits_read_image_speed (outfptr, &whole_elapse, &whole_cpu, 
 	   &row_elapse, &row_cpu, &stat);
-	
+	}
 
         if (!stat) {
 
 		/* -------------- UNCOMPRESS the image ------------------ */
+
+                /* remove any cached uncompressed tile 
+		  (dangerous to directly modify the structure!) */
+                (outfptr->Fptr)->tilerow = 0;
 		marktime(&stat);
 
  	       fits_img_decompress (outfptr, outfptr2, &stat);
@@ -1719,5 +1997,5 @@ void abort_fpack(int sig)
       if (tempfilename3[0]) {
          remove(tempfilename3);
       }
-      exit(-1); 
+      exit(-1);
 }
